from enum import Enum

from roborock.code_mappings import STATE_CODE_TO_STATUS, ERROR_CODE_TO_TEXT, FAN_SPEED_CODES, MOP_MODE_CODES, \
    MOP_INTENSITY_CODES


class UserDataRRiotReferenceField(str, Enum):
    REGION = "r"
    API = "a"
    MQTT = "m"
    L_UNKNOWN = "l"


class UserDataRRiotField(str, Enum):
    USER = "u"
    PASSWORD = "s"
    H_UNKNOWN = "h"
    DOMAIN = "k"
    REFERENCE = "r"


class UserDataField(str, Enum):
    UID = "uid"
    TOKEN_TYPE = "tokentype"
    TOKEN = "token"
    RR_UID = "rruid"
    REGION = "region"
    COUNTRY_CODE = "countrycode"
    COUNTRY = "country"
    NICKNAME = "nickname"
    RRIOT = "rriot"
    TUYA_DEVICE_STATE = "tuyaDeviceState"
    AVATAR_URL = "avatarurl"


class HomeDataProductSchemaField(str, Enum):
    ID = "id"
    NAME = "name"
    CODE = "code"
    MODE = "mode"
    TYPE = "type"
    PROPERTY = "property"
    DESC = "desc"


class HomeDataProductField(str, Enum):
    ID = "id"
    NAME = "name"
    CODE = "code"
    MODEL = "model"
    ICONURL = "iconUrl"
    ATTRIBUTE = "attribute"
    CAPABILITY = "capability"
    CATEGORY = "category"
    SCHEMA = "schema"


class HomeDataDeviceStatusField(str, Enum):
    ID = "id"
    NAME = "name"
    CODE = "code"
    MODEL = "model"
    ICON_URL = "iconUrl"
    ATTRIBUTE = "attribute"
    CAPABILITY = "capability"
    CATEGORY = "category"
    SCHEMA = "schema"


class HomeDataDeviceField(str, Enum):
    DUID = "duid"
    NAME = "name"
    ATTRIBUTE = "attribute"
    ACTIVETIME = "activeTime"
    LOCAL_KEY = "localKey"
    RUNTIME_ENV = "runtimeEnv"
    TIME_ZONE_ID = "timeZoneId"
    ICON_URL = "iconUrl"
    PRODUCT_ID = "productId"
    LON = "lon"
    LAT = "lat"
    SHARE = "share"
    SHARE_TIME = "shareTime"
    ONLINE = "online"
    FV = "fv"
    PV = "pv"
    ROOM_ID = "roomId"
    TUYA_UUID = "tuyaUuid"
    TUYA_MIGRATED = "tuyaMigrated"
    EXTRA = "extra"
    SN = "sn"
    FEATURE_SET = "featureSet"
    NEW_FEATURE_SET = "newFeatureSet"
    DEVICE_STATUS = "deviceStatus"
    SILENT_OTA_SWITCH = "silentOtaSwitch"


class HomeDataRoomField(str, Enum):
    ID = "id"
    NAME = "name"


class HomeDataField(str, Enum):
    ID = "id"
    NAME = "name"
    LON = "lon"
    LAT = "lat"
    GEO_NAME = "geoName"
    PRODUCTS = "products"
    DEVICES = "devices"
    RECEIVED_DEVICES = "receivedDevices"
    ROOMS = "rooms"


class StatusField(str, Enum):
    MSG_VER = "msg_ver"
    MSG_SEQ = "msg_seq"
    STATE = "state"
    BATTERY = "battery"
    CLEAN_TIME = "clean_time"
    CLEAN_AREA = "clean_area"
    ERROR_CODE = "error_code"
    MAP_PRESENT = "map_present"
    IN_CLEANING = "in_cleaning"
    IN_RETURNING = "in_returning"
    IN_FRESH_STATE = "in_fresh_state"
    LAB_STATUS = "lab_status"
    WATER_BOX_STATUS = "water_box_status"
    BACK_TYPE = "back_type"
    WASH_PHASE = "wash_phase"
    WASH_READY = "wash_ready"
    FAN_POWER = "fan_power"
    DND_ENABLED = "dnd_enabled"
    MAP_STATUS = "map_status"
    IS_LOCATING = "is_locating"
    LOCK_STATUS = "lock_status"
    WATER_BOX_MODE = "water_box_mode"
    WATER_BOX_CARRIAGE_STATUS = "water_box_carriage_status"
    MOP_FORBIDDEN_ENABLE = "mop_forbidden_enable"
    CAMERA_STATUS = "camera_status"
    IS_EXPLORING = "is_exploring"
    HOME_SEC_STATUS = "home_sec_status"
    HOME_SEC_ENABLE_PASSWORD = "home_sec_enable_password"
    ADBUMPER_STATUS = "adbumper_status"
    WATER_SHORTAGE_STATUS = "water_shortage_status"
    DOCK_TYPE = "dock_type"
    DUST_COLLECTION_STATUS = "dust_collection_status"
    AUTO_DUST_COLLECTION = "auto_dust_collection"
    AVOID_COUNT = "avoid_count"
    MOP_MODE = "mop_mode"
    DEBUG_MODE = "debug_mode"
    COLLISION_AVOID_STATUS = "collision_avoid_status"
    SWITCH_MAP_MODE = "switch_map_mode"
    DOCK_ERROR_STATUS = "dock_error_status"
    CHARGE_STATUS = "charge_status"
    UNSAVE_MAP_REASON = "unsave_map_reason"
    UNSAVE_MAP_FLAG = "unsave_map_flag"


class DNDTimerField(str, Enum):
    START_HOUR = "start_hour"
    START_MINUTE = "start_minute"
    END_HOUR = "end_hour"
    END_MINUTE = "end_minute"
    ENABLED = "enabled"


class CleanSummaryField(str, Enum):
    CLEAN_TIME = "clean_time"
    CLEAN_AREA = "clean_area"
    CLEAN_COUNT = "clean_count"
    DUST_COLLECTION_COUNT = "dust_collection_count"
    RECORDS = "records"


class CleanRecordField(str, Enum):
    BEGIN = "begin"
    END = "end"
    DURATION = "duration"
    AREA = "area"
    ERROR = "error"
    COMPLETE = "complete"
    START_TYPE = "start_type"
    CLEAN_TYPE = "clean_type"
    FINISH_REASON = "finish_reason"
    DUST_COLLECTION_STATUS = "dust_collection_status"
    AVOID_COUNT = "avoid_count"
    WASH_COUNT = "wash_count"
    MAP_FLAG = "map_flag"


class ConsumableField(str, Enum):
    MAIN_BRUSH_WORK_TIME = "main_brush_work_time"
    SIDE_BRUSH_WORK_TIME = "side_brush_work_time"
    FILTER_WORK_TIME = "filter_work_time"
    FILTER_ELEMENT_WORK_TIME = "filter_element_work_time"
    SENSOR_DIRTY_TIME = "sensor_dirty_time"
    STRAINER_WORK_TIMES = "strainer_work_times"
    DUST_COLLECTION_WORK_TIMES = "dust_collection_work_times"
    CLEANING_BRUSH_WORK_TIMES = "cleaning_brush_work_times"


class MultiMapListMapInfoBakMapsField(str, Enum):
    MAPFLAG = "mapFlag"
    ADD_TIME = "add_time"


class MultiMapListMapInfoField(str, Enum):
    MAPFLAG = "mapFlag"
    ADD_TIME = "add_time"
    LENGTH = "length"
    NAME = "name"
    BAK_MAPS = "bak_maps"


class MultiMapListField(str, Enum):
    MAX_MULTI_MAP = "max_multi_map"
    MAX_BAK_MAP = "max_bak_map"
    MULTI_MAP_COUNT = "multi_map_count"
    MAP_INFO = "map_info"


class RoborockBase(dict):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__()
        if isinstance(data, dict):
            self.update(data)

    def as_dict(self):
        return self.__dict__


class Reference(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def region(self) -> str:
        return self.get(UserDataRRiotReferenceField.REGION)

    @property
    def api(self) -> str:
        return self.get(UserDataRRiotReferenceField.API)

    @property
    def mqtt(self) -> str:
        return self.get(UserDataRRiotReferenceField.MQTT)

    @property
    def l_unknown(self) -> str:
        return self.get(UserDataRRiotReferenceField.L_UNKNOWN)


class RRiot(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def user(self) -> str:
        return self.get(UserDataRRiotField.USER)

    @property
    def password(self) -> str:
        return self.get(UserDataRRiotField.PASSWORD)

    @property
    def h_unknown(self) -> str:
        return self.get(UserDataRRiotField.H_UNKNOWN)

    @property
    def domain(self) -> str:
        return self.get(UserDataRRiotField.DOMAIN)

    @property
    def reference(self) -> Reference:
        return Reference(self.get(UserDataRRiotField.REFERENCE))


class UserData(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def uid(self) -> int:
        return self.get(UserDataField.UID)

    @property
    def token_type(self) -> str:
        return self.get(UserDataField.TOKEN_TYPE)

    @property
    def token(self) -> str:
        return self.get(UserDataField.TOKEN)

    @property
    def rr_uid(self) -> str:
        return self.get(UserDataField.RR_UID)

    @property
    def region(self) -> str:
        return self.get(UserDataField.REGION)

    @property
    def country_code(self) -> str:
        return self.get(UserDataField.COUNTRY_CODE)

    @property
    def country(self) -> str:
        return self.get(UserDataField.COUNTRY)

    @property
    def nickname(self) -> str:
        return self.get(UserDataField.NICKNAME)

    @property
    def rriot(self) -> RRiot:
        return RRiot(self.get(UserDataField.RRIOT))

    @property
    def tuya_device_state(self) -> int:
        return self.get(UserDataField.TUYA_DEVICE_STATE)

    @property
    def avatar_url(self) -> str:
        return self.get(UserDataField.AVATAR_URL)


class LoginData(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def user_data(self) -> UserData:
        user_data = self.get("user_data")
        if user_data:
            return UserData(user_data)

    @property
    def home_data(self):
        home_data = self.get("home_data")
        if home_data:
            return HomeData(home_data)

    @property
    def email(self):
        return self.get("email")


class HomeDataProductSchema(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def id(self):
        return self.get(HomeDataProductSchemaField.ID)

    @property
    def name(self):
        return self.get(HomeDataProductSchemaField.NAME)

    @property
    def code(self):
        return self.get(HomeDataProductSchemaField.CODE)

    @property
    def mode(self):
        return self.get(HomeDataProductSchemaField.MODE)

    @property
    def type(self):
        return self.get(HomeDataProductSchemaField.TYPE)

    @property
    def product_property(self):
        return self.get(HomeDataProductSchemaField.PROPERTY)

    @property
    def desc(self):
        return self.get(HomeDataProductSchemaField.DESC)


class HomeDataProduct(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def id(self) -> str:
        return self.get(HomeDataProductField.ID)

    @property
    def name(self) -> str:
        return self.get(HomeDataProductField.NAME)

    @property
    def code(self) -> str:
        return self.get(HomeDataProductField.CODE)

    @property
    def model(self) -> str:
        return self.get(HomeDataProductField.MODEL)

    @property
    def iconurl(self) -> str:
        return self.get(HomeDataProductField.ICONURL)

    @property
    def attribute(self):
        return self.get(HomeDataProductField.ATTRIBUTE)

    @property
    def capability(self) -> int:
        return self.get(HomeDataProductField.CAPABILITY)

    @property
    def category(self) -> str:
        return self.get(HomeDataProductField.CATEGORY)

    @property
    def schema(self) -> list[HomeDataProductSchema]:
        return [HomeDataProductSchema(schema) for schema in self.get(HomeDataProductField.SCHEMA)]


class HomeDataDeviceStatus(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def id(self):
        return self.get(HomeDataDeviceStatusField.ID)

    @property
    def name(self):
        return self.get(HomeDataDeviceStatusField.NAME)

    @property
    def code(self):
        return self.get(HomeDataDeviceStatusField.CODE)

    @property
    def model(self):
        return self.get(HomeDataDeviceStatusField.MODEL)

    @property
    def icon_url(self):
        return self.get(HomeDataDeviceStatusField.ICON_URL)

    @property
    def attribute(self):
        return self.get(HomeDataDeviceStatusField.ATTRIBUTE)

    @property
    def capability(self):
        return self.get(HomeDataDeviceStatusField.CAPABILITY)

    @property
    def category(self):
        return self.get(HomeDataDeviceStatusField.CATEGORY)

    @property
    def schema(self):
        return self.get(HomeDataDeviceStatusField.SCHEMA)


class HomeDataDevice(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def duid(self) -> str:
        return self.get(HomeDataDeviceField.DUID)

    @property
    def name(self) -> str:
        return self.get(HomeDataDeviceField.NAME)

    @property
    def attribute(self):
        return self.get(HomeDataDeviceField.ATTRIBUTE)

    @property
    def activetime(self) -> int:
        return self.get(HomeDataDeviceField.ACTIVETIME)

    @property
    def local_key(self) -> str:
        return self.get(HomeDataDeviceField.LOCAL_KEY)

    @property
    def runtime_env(self):
        return self.get(HomeDataDeviceField.RUNTIME_ENV)

    @property
    def time_zone_id(self) -> str:
        return self.get(HomeDataDeviceField.TIME_ZONE_ID)

    @property
    def icon_url(self) -> str:
        return self.get(HomeDataDeviceField.ICON_URL)

    @property
    def product_id(self) -> str:
        return self.get(HomeDataDeviceField.PRODUCT_ID)

    @property
    def lon(self):
        return self.get(HomeDataDeviceField.LON)

    @property
    def lat(self):
        return self.get(HomeDataDeviceField.LAT)

    @property
    def share(self) -> bool:
        return self.get(HomeDataDeviceField.SHARE)

    @property
    def share_time(self):
        return self.get(HomeDataDeviceField.SHARE_TIME)

    @property
    def online(self) -> bool:
        return self.get(HomeDataDeviceField.ONLINE)

    @property
    def fv(self) -> str:
        return self.get(HomeDataDeviceField.FV)

    @property
    def pv(self) -> str:
        return self.get(HomeDataDeviceField.PV)

    @property
    def room_id(self):
        return self.get(HomeDataDeviceField.ROOM_ID)

    @property
    def tuya_uuid(self):
        return self.get(HomeDataDeviceField.TUYA_UUID)

    @property
    def tuya_migrated(self) -> bool:
        return self.get(HomeDataDeviceField.TUYA_MIGRATED)

    @property
    def extra(self):
        return self.get(HomeDataDeviceField.EXTRA)

    @property
    def sn(self) -> str:
        return self.get(HomeDataDeviceField.SN)

    @property
    def feature_set(self) -> str:
        return self.get(HomeDataDeviceField.FEATURE_SET)

    @property
    def new_feature_set(self) -> str:
        return self.get(HomeDataDeviceField.NEW_FEATURE_SET)

    @property
    def device_status(self) -> HomeDataDeviceStatus:
        return HomeDataDeviceStatus(self.get(HomeDataDeviceField.DEVICE_STATUS))

    @property
    def silent_ota_switch(self) -> bool:
        return self.get(HomeDataDeviceField.SILENT_OTA_SWITCH)


class HomeDataRoom(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def id(self):
        return self.get(HomeDataRoomField.ID)

    @property
    def name(self):
        return self.get(HomeDataRoomField.NAME)


class HomeData(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def id(self) -> int:
        return self.get(HomeDataField.ID)

    @property
    def name(self) -> str:
        return self.get(HomeDataField.NAME)

    @property
    def lon(self):
        return self.get(HomeDataField.LON)

    @property
    def lat(self):
        return self.get(HomeDataField.LAT)

    @property
    def geo_name(self):
        return self.get(HomeDataField.GEO_NAME)

    @property
    def products(self) -> list[HomeDataProduct]:
        return [HomeDataProduct(product) for product in self.get(HomeDataField.PRODUCTS)]

    @property
    def devices(self) -> list[HomeDataDevice]:
        return [HomeDataDevice(device) for device in self.get(HomeDataField.DEVICES)]

    @property
    def received_devices(self) -> list[HomeDataDevice]:
        return [HomeDataDevice(device) for device in self.get(HomeDataField.RECEIVED_DEVICES)]

    @property
    def rooms(self) -> list[HomeDataRoom]:
        return [HomeDataRoom(room) for room in self.get(HomeDataField.ROOMS)]


class Status(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def msg_ver(self) -> int:
        return self.get(StatusField.MSG_VER)

    @property
    def msg_seq(self) -> int:
        return self.get(StatusField.MSG_SEQ)

    @property
    def status(self) -> str:
        return STATE_CODE_TO_STATUS.get(self.state)

    @property
    def state(self) -> int:
        return self.get(StatusField.STATE)

    @property
    def battery(self) -> int:
        return self.get(StatusField.BATTERY)

    @property
    def clean_time(self) -> int:
        return self.get(StatusField.CLEAN_TIME)

    @property
    def clean_area(self) -> int:
        return self.get(StatusField.CLEAN_AREA)

    @property
    def error_code(self) -> int:
        return self.get(StatusField.ERROR_CODE)

    @property
    def error(self) -> str:
        return ERROR_CODE_TO_TEXT.get(self.error_code)

    @property
    def map_present(self) -> int:
        return self.get(StatusField.MAP_PRESENT)

    @property
    def in_cleaning(self) -> int:
        return self.get(StatusField.IN_CLEANING)

    @property
    def in_returning(self) -> int:
        return self.get(StatusField.IN_RETURNING)

    @property
    def in_fresh_state(self) -> int:
        return self.get(StatusField.IN_FRESH_STATE)

    @property
    def lab_status(self) -> int:
        return self.get(StatusField.LAB_STATUS)

    @property
    def water_box_status(self) -> int:
        return self.get(StatusField.WATER_BOX_STATUS)

    @property
    def back_type(self) -> int:
        return self.get(StatusField.BACK_TYPE)

    @property
    def wash_phase(self) -> int:
        return self.get(StatusField.WASH_PHASE)

    @property
    def wash_ready(self) -> int:
        return self.get(StatusField.WASH_READY)

    @property
    def fan_power_code(self) -> int:
        return self.get(StatusField.FAN_POWER)

    @property
    def fan_power(self) -> str:
        return FAN_SPEED_CODES.get(self.fan_power_code)

    @property
    def dnd_enabled(self) -> int:
        return self.get(StatusField.DND_ENABLED)

    @property
    def map_status(self) -> int:
        return self.get(StatusField.MAP_STATUS)

    @property
    def is_locating(self) -> int:
        return self.get(StatusField.IS_LOCATING)

    @property
    def lock_status(self) -> int:
        return self.get(StatusField.LOCK_STATUS)

    @property
    def water_box_mode(self) -> int:
        return self.get(StatusField.WATER_BOX_MODE)

    @property
    def mop_intensity(self) -> str:
        return MOP_INTENSITY_CODES.get(self.water_box_mode)

    @property
    def water_box_carriage_status(self) -> int:
        return self.get(StatusField.WATER_BOX_CARRIAGE_STATUS)

    @property
    def mop_forbidden_enable(self) -> int:
        return self.get(StatusField.MOP_FORBIDDEN_ENABLE)

    @property
    def camera_status(self) -> int:
        return self.get(StatusField.CAMERA_STATUS)

    @property
    def is_exploring(self) -> int:
        return self.get(StatusField.IS_EXPLORING)

    @property
    def home_sec_status(self) -> int:
        return self.get(StatusField.HOME_SEC_STATUS)

    @property
    def home_sec_enable_password(self) -> int:
        return self.get(StatusField.HOME_SEC_ENABLE_PASSWORD)

    @property
    def adbumper_status(self) -> list[int]:
        return self.get(StatusField.ADBUMPER_STATUS)

    @property
    def water_shortage_status(self) -> int:
        return self.get(StatusField.WATER_SHORTAGE_STATUS)

    @property
    def dock_type(self) -> int:
        return self.get(StatusField.DOCK_TYPE)

    @property
    def dust_collection_status(self) -> int:
        return self.get(StatusField.DUST_COLLECTION_STATUS)

    @property
    def auto_dust_collection(self) -> int:
        return self.get(StatusField.AUTO_DUST_COLLECTION)

    @property
    def avoid_count(self) -> int:
        return self.get(StatusField.AVOID_COUNT)

    @property
    def mop_mode_code(self) -> int:
        return self.get(StatusField.MOP_MODE)

    @property
    def mop_mode(self) -> str:
        return MOP_MODE_CODES.get(self.mop_mode_code)

    @property
    def debug_mode(self) -> int:
        return self.get(StatusField.DEBUG_MODE)

    @property
    def collision_avoid_status(self) -> int:
        return self.get(StatusField.COLLISION_AVOID_STATUS)

    @property
    def switch_map_mode(self) -> int:
        return self.get(StatusField.SWITCH_MAP_MODE)

    @property
    def dock_error_status(self) -> int:
        return self.get(StatusField.DOCK_ERROR_STATUS)

    @property
    def charge_status(self) -> int:
        return self.get(StatusField.CHARGE_STATUS)

    @property
    def unsave_map_reason(self) -> int:
        return self.get(StatusField.UNSAVE_MAP_REASON)

    @property
    def unsave_map_flag(self) -> int:
        return self.get(StatusField.UNSAVE_MAP_FLAG)


class DNDTimer(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def start_hour(self) -> int:
        return self.get(DNDTimerField.START_HOUR)

    @property
    def start_minute(self) -> int:
        return self.get(DNDTimerField.START_MINUTE)

    @property
    def end_hour(self) -> int:
        return self.get(DNDTimerField.END_HOUR)

    @property
    def end_minute(self) -> int:
        return self.get(DNDTimerField.END_MINUTE)

    @property
    def enabled(self) -> int:
        return self.get(DNDTimerField.ENABLED)


class CleanSummary(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def clean_time(self) -> int:
        return self.get(CleanSummaryField.CLEAN_TIME)

    @property
    def clean_area(self) -> int:
        return self.get(CleanSummaryField.CLEAN_AREA)

    @property
    def clean_count(self) -> int:
        return self.get(CleanSummaryField.CLEAN_COUNT)

    @property
    def dust_collection_count(self) -> int:
        return self.get(CleanSummaryField.DUST_COLLECTION_COUNT)

    @property
    def records(self) -> list[int]:
        return self.get(CleanSummaryField.RECORDS)


class CleanRecord(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def begin(self) -> int:
        return self.get(CleanRecordField.BEGIN)

    @property
    def end(self) -> int:
        return self.get(CleanRecordField.END)

    @property
    def duration(self) -> int:
        return self.get(CleanRecordField.DURATION)

    @property
    def area(self) -> int:
        return self.get(CleanRecordField.AREA)

    @property
    def error(self) -> int:
        return self.get(CleanRecordField.ERROR)

    @property
    def complete(self) -> int:
        return self.get(CleanRecordField.COMPLETE)

    @property
    def start_type(self) -> int:
        return self.get(CleanRecordField.START_TYPE)

    @property
    def clean_type(self) -> int:
        return self.get(CleanRecordField.CLEAN_TYPE)

    @property
    def finish_reason(self) -> int:
        return self.get(CleanRecordField.FINISH_REASON)

    @property
    def dust_collection_status(self) -> int:
        return self.get(CleanRecordField.DUST_COLLECTION_STATUS)

    @property
    def avoid_count(self) -> int:
        return self.get(CleanRecordField.AVOID_COUNT)

    @property
    def wash_count(self) -> int:
        return self.get(CleanRecordField.WASH_COUNT)

    @property
    def map_flag(self) -> int:
        return self.get(CleanRecordField.MAP_FLAG)


class Consumable(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def main_brush_work_time(self) -> int:
        return self.get(ConsumableField.MAIN_BRUSH_WORK_TIME)

    @property
    def side_brush_work_time(self) -> int:
        return self.get(ConsumableField.SIDE_BRUSH_WORK_TIME)

    @property
    def filter_work_time(self) -> int:
        return self.get(ConsumableField.FILTER_WORK_TIME)

    @property
    def filter_element_work_time(self) -> int:
        return self.get(ConsumableField.FILTER_ELEMENT_WORK_TIME)

    @property
    def sensor_dirty_time(self) -> int:
        return self.get(ConsumableField.SENSOR_DIRTY_TIME)

    @property
    def strainer_work_times(self) -> int:
        return self.get(ConsumableField.STRAINER_WORK_TIMES)

    @property
    def dust_collection_work_times(self) -> int:
        return self.get(ConsumableField.DUST_COLLECTION_WORK_TIMES)

    @property
    def cleaning_brush_work_times(self) -> int:
        return self.get(ConsumableField.CLEANING_BRUSH_WORK_TIMES)


class MultiMapsListMapInfoBakMaps(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def mapflag(self):
        return self.get(MultiMapListMapInfoBakMapsField.MAPFLAG)

    @property
    def add_time(self):
        return self.get(MultiMapListMapInfoBakMapsField.ADD_TIME)


class MultiMapsListMapInfo(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def mapflag(self):
        return self.get(MultiMapListMapInfoField.MAPFLAG)

    @property
    def add_time(self):
        return self.get(MultiMapListMapInfoField.ADD_TIME)

    @property
    def length(self):
        return self.get(MultiMapListMapInfoField.LENGTH)

    @property
    def name(self):
        return self.get(MultiMapListMapInfoField.NAME)

    @property
    def bak_maps(self):
        return [MultiMapsListMapInfoBakMaps(bak_maps) for bak_maps in self.get(MultiMapListMapInfoField.BAK_MAPS)]


class MultiMapsList(RoborockBase):
    def __init__(self, data: dict[str, any]) -> None:
        super().__init__(data)

    @property
    def max_multi_map(self) -> int:
        return self.get(MultiMapListField.MAX_MULTI_MAP)

    @property
    def max_bak_map(self) -> int:
        return self.get(MultiMapListField.MAX_BAK_MAP)

    @property
    def multi_map_count(self) -> int:
        return self.get(MultiMapListField.MULTI_MAP_COUNT)

    @property
    def map_info(self) -> list[MultiMapsListMapInfo]:
        return [MultiMapsListMapInfo(map_info) for map_info in self.get(MultiMapListField.MAP_INFO)]



import json
import logging
from pathlib import Path
from typing import Any, Dict

import click

from roborock import RoborockException
from roborock.api import RoborockClient, RoborockMqttClient
from roborock.containers import LoginData
from roborock.typing import RoborockDeviceInfo
from roborock.util import run_sync

_LOGGER = logging.getLogger(__name__)


class RoborockContext:
    roborock_file = Path("~/.roborock").expanduser()
    _login_data: LoginData = None

    def __init__(self):
        self.reload()

    def reload(self):
        if self.roborock_file.is_file():
            with open(self.roborock_file, 'r') as f:
                data = json.load(f)
                if data:
                    self._login_data = LoginData(data)

    def update(self, login_data: LoginData):
        data = json.dumps(login_data, default=vars)
        with open(self.roborock_file, 'w') as f:
            f.write(data)
        self.reload()

    def validate(self):
        if self._login_data is None:
            raise RoborockException("You must login first")

    def login_data(self):
        self.validate()
        return self._login_data


@click.option("-d", "--debug", default=False, count=True)
@click.version_option(package_name="python-roborock")
@click.group()
@click.pass_context
def cli(ctx, debug: int):
    logging_config: Dict[str, Any] = {
        "level": logging.DEBUG if debug > 0 else logging.INFO
    }
    logging.basicConfig(**logging_config)  # type: ignore
    ctx.obj = RoborockContext()


@click.command()
@click.option("--email", required=True)
@click.option("--password", required=True)
@click.pass_context
@run_sync()
async def login(ctx, email, password):
    """Login to Roborock account."""
    context: RoborockContext = ctx.obj
    try:
        context.validate()
        _LOGGER.info("Already logged in")
        return
    except RoborockException:
        pass
    client = RoborockClient(email)
    user_data = await client.pass_login(password)
    context.update(LoginData({"user_data": user_data, "email": email}))

async def _discover(ctx):
    context: RoborockContext = ctx.obj
    login_data = context.login_data()
    client = RoborockClient(login_data.email)
    home_data = await client.get_home_data(login_data.user_data)
    context.update(LoginData({**login_data, "home_data": home_data}))
    click.echo(f"Discovered devices {', '.join([device.name for device in home_data.devices + home_data.received_devices])}")

@click.command()
@click.pass_context
@run_sync()
async def discover(ctx):
    await _discover(ctx)

@click.command()
@click.pass_context
@run_sync()
async def list_devices(ctx):
    context: RoborockContext = ctx.obj
    login_data = context.login_data()
    if not login_data.home_data:
        await _discover(ctx)
        login_data = context.login_data()
    home_data = login_data.home_data
    click.echo(f"Known devices {', '.join([device.name for device in home_data.devices + home_data.received_devices])}")

@click.command()
@click.pass_context
@run_sync()
async def get_status(ctx):
    context: RoborockContext = ctx.obj
    login_data = context.login_data()
    if not login_data.home_data:
        await _discover(ctx)
        login_data = context.login_data()
    home_data = login_data.home_data
    device_map: dict[str, RoborockDeviceInfo] = {}
    for device in home_data.devices + home_data.received_devices:
        product = next(
            (
                product
                for product in home_data.products
                if product.id == device.product_id
            ),
            {},
        )
        device_map[device.duid] = RoborockDeviceInfo(device, product)
    mqtt_client = RoborockMqttClient(login_data.user_data, device_map)
    await mqtt_client.get_status(home_data.devices[0].duid)
    mqtt_client.__del__()


cli.add_command(login)
cli.add_command(discover)
cli.add_command(list_devices)
cli.add_command(get_status)


def main():
    return cli()


if __name__ == "__main__":
    main()



"""The Roborock api."""
from __future__ import annotations

import asyncio
import base64
import binascii
import gzip
import hashlib
import hmac
import json
import logging
import math
import secrets
import struct
import threading
import time
from asyncio import Lock
from asyncio.exceptions import TimeoutError, CancelledError
from typing import Any
from urllib.parse import urlparse

import aiohttp
import paho.mqtt.client as mqtt
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad

from roborock.containers import (
    UserData,
    HomeDataDevice,
    Status,
    CleanSummary,
    Consumable,
    DNDTimer,
    CleanRecord,
    HomeData,
    MultiMapsList,
)
from roborock.exceptions import (
    RoborockException,
    CommandVacuumError,
    VacuumError,
    RoborockTimeout,
)
from roborock.roborock_queue import RoborockQueue
from roborock.typing import (
    RoborockDeviceInfo,
    RoborockDeviceProp,
    RoborockCommand,
)
from roborock.util import run_in_executor

_LOGGER = logging.getLogger(__name__)
QUEUE_TIMEOUT = 4
MQTT_KEEPALIVE = 60


def md5hex(message: str) -> str:
    md5 = hashlib.md5()
    md5.update(message.encode())
    return md5.hexdigest()


def md5bin(message: str) -> bytes:
    md5 = hashlib.md5()
    md5.update(message.encode())
    return md5.digest()


def encode_timestamp(_timestamp: int) -> str:
    hex_value = f"{_timestamp:x}".zfill(8)
    return "".join(list(map(lambda idx: hex_value[idx], [5, 6, 3, 7, 1, 2, 0, 4])))


class PreparedRequest:
    def __init__(self, base_url: str, base_headers: dict = None) -> None:
        self.base_url = base_url
        self.base_headers = base_headers or {}

    async def request(
            self, method: str, url: str, params=None, data=None, headers=None
    ) -> dict | list:
        _url = "/".join(s.strip("/") for s in [self.base_url, url])
        _headers = {**self.base_headers, **(headers or {})}
        async with aiohttp.ClientSession() as session:
            async with session.request(
                    method,
                    _url,
                    params=params,
                    data=data,
                    headers=_headers,
            ) as resp:
                return await resp.json()


COMMANDS_WITH_BINARY_RESPONSE = [
    RoborockCommand.GET_MAP_V1,
]


class RoborockMqttClient(mqtt.Client):
    _thread: threading.Thread

    def __init__(self, user_data: UserData, device_map: dict[str, RoborockDeviceInfo]) -> None:
        rriot = user_data.rriot
        self._mqtt_user = rriot.user
        self._mqtt_domain = rriot.domain
        self._hashed_user = md5hex(self._mqtt_user + ":" + self._mqtt_domain)[2:10]
        super().__init__(client_id=self._hashed_user, protocol=mqtt.MQTTv5)
        url = urlparse(rriot.reference.mqtt)
        self._mqtt_host = url.hostname
        self._mqtt_port = url.port
        self._mqtt_ssl = url.scheme == "ssl"
        if self._mqtt_ssl:
            super().tls_set()
        self._mqtt_password = rriot.password
        self._hashed_password = md5hex(self._mqtt_password + ":" + self._mqtt_domain)[16:]
        super().username_pw_set(self._hashed_user, self._hashed_password)
        self.device_map = device_map
        self._seq = 1
        self._random = 4711
        self._id_counter = 2
        self._salt = "TXdfu$jyZ#TZHsg4"
        self._endpoint = base64.b64encode(md5bin(self._mqtt_domain)[8:14]).decode()
        self._nonce = secrets.token_bytes(16)
        self._waiting_queue: dict[int, RoborockQueue] = {}
        self._mutex = Lock()
        self._last_device_msg_in = mqtt.time_func()
        self._last_disconnection = mqtt.time_func()

    def __del__(self) -> None:
        self.sync_disconnect()

    @run_in_executor()
    async def on_connect(self, _client, _, __, rc, ___=None) -> None:
        connection_queue = self._waiting_queue.get(0)
        if rc != mqtt.MQTT_ERR_SUCCESS:
            message = f"Failed to connect (rc: {rc})"
            _LOGGER.error(message)
            if connection_queue:
                await connection_queue.async_put(
                    (None, VacuumError(rc, message)), timeout=QUEUE_TIMEOUT
                )
            return
        _LOGGER.info(f"Connected to mqtt {self._mqtt_host}:{self._mqtt_port}")
        topic = f"rr/m/o/{self._mqtt_user}/{self._hashed_user}/#"
        (result, mid) = self.subscribe(topic)
        if result != 0:
            message = f"Failed to subscribe (rc: {result})"
            _LOGGER.error(message)
            if connection_queue:
                await connection_queue.async_put(
                    (None, VacuumError(rc, message)), timeout=QUEUE_TIMEOUT
                )
            return
        _LOGGER.info(f"Subscribed to topic {topic}")
        if connection_queue:
            await connection_queue.async_put((True, None), timeout=QUEUE_TIMEOUT)

    @run_in_executor()
    async def on_message(self, _client, _, msg, __=None) -> None:
        try:
            async with self._mutex:
                self._last_device_msg_in = mqtt.time_func()
            device_id = msg.topic.split("/").pop()
            data = self._decode_msg(msg.payload, self.device_map[device_id].device)
            protocol = data.get("protocol")
            if protocol == 102:
                payload = json.loads(data.get("payload").decode())
                for data_point_number, data_point in payload.get("dps").items():
                    if data_point_number == "102":
                        data_point_response = json.loads(data_point)
                        request_id = data_point_response.get("id")
                        queue = self._waiting_queue.get(request_id)
                        if queue:
                            if queue.protocol == protocol:
                                error = data_point_response.get("error")
                                if error:
                                    await queue.async_put(
                                        (
                                            None,
                                            VacuumError(
                                                error.get("code"), error.get("message")
                                            ),
                                        ),
                                        timeout=QUEUE_TIMEOUT,
                                    )
                                else:
                                    result = data_point_response.get("result")
                                    if isinstance(result, list) and len(result) > 0:
                                        result = result[0]
                                    await queue.async_put(
                                        (result, None), timeout=QUEUE_TIMEOUT
                                    )
                        elif request_id < self._id_counter:
                            _LOGGER.debug(
                                f"id={request_id} Ignoring response: {data_point_response}"
                            )
                    elif data_point_number == "121":
                        _LOGGER.debug(f"Remote control {data_point}")
                    else:
                        _LOGGER.debug(
                            f"Unknown data point number received {data_point_number} with {data_point}"
                        )
            elif protocol == 301:
                payload = data.get("payload")[0:24]
                [endpoint, _, request_id, _] = struct.unpack("<15sBH6s", payload)
                if endpoint.decode().startswith(self._endpoint):
                    iv = bytes(AES.block_size)
                    decipher = AES.new(self._nonce, AES.MODE_CBC, iv)
                    decrypted = unpad(
                        decipher.decrypt(data.get("payload")[24:]), AES.block_size
                    )
                    decrypted = gzip.decompress(decrypted)
                    queue = self._waiting_queue.get(request_id)
                    if queue:
                        if isinstance(decrypted, list):
                            decrypted = decrypted[0]
                        await queue.async_put((decrypted, None), timeout=QUEUE_TIMEOUT)
        except Exception as ex:
            _LOGGER.exception(ex)

    @run_in_executor()
    async def on_disconnect(self, _client: mqtt.Client, _, rc, __=None) -> None:
        try:
            async with self._mutex:
                self._last_disconnection = mqtt.time_func()
            message = f"Roborock mqtt client disconnected (rc: {rc})"
            _LOGGER.warning(message)
            connection_queue = self._waiting_queue.get(1)
            if rc != mqtt.MQTT_ERR_SUCCESS and rc != mqtt.MQTT_ERR_KEEPALIVE:
                await self.async_disconnect()
                if connection_queue:
                    await connection_queue.async_put(
                        (None, VacuumError(rc, message)), timeout=QUEUE_TIMEOUT
                    )
                return
            if connection_queue:
                await connection_queue.async_put(
                    (True, None), timeout=QUEUE_TIMEOUT
                )
        except Exception as ex:
            _LOGGER.exception(ex)

    @run_in_executor()
    async def _async_check_keepalive(self) -> None:
        async with self._mutex:
            now = mqtt.time_func()
            if now - self._last_disconnection > self._keepalive ** 2 and now - self._last_device_msg_in > self._keepalive:
                self._ping_t = self._last_device_msg_in

    def _check_keepalive(self) -> None:
        self._async_check_keepalive()
        super()._check_keepalive()

    def sync_stop_loop(self) -> None:
        if self._thread:
            _LOGGER.info("Stopping mqtt loop")
            super().loop_stop()

    def sync_start_loop(self) -> None:
        if not self._thread or not self._thread.is_alive():
            self.sync_stop_loop()
            _LOGGER.info("Starting mqtt loop")
            super().loop_start()

    def sync_disconnect(self) -> bool:
        rc = mqtt.MQTT_ERR_AGAIN
        if self.is_connected():
            _LOGGER.info("Disconnecting from mqtt")
            rc = super().disconnect()
            if not rc in [mqtt.MQTT_ERR_SUCCESS, mqtt.MQTT_ERR_NO_CONN]:
                raise RoborockException(f"Failed to disconnect (rc:{rc})")
        return rc == mqtt.MQTT_ERR_SUCCESS

    def sync_connect(self) -> bool:
        rc = mqtt.MQTT_ERR_AGAIN
        self.sync_start_loop()
        if not self.is_connected():
            _LOGGER.info("Connecting to mqtt")
            rc = super().connect(
                host=self._mqtt_host,
                port=self._mqtt_port,
                keepalive=MQTT_KEEPALIVE
            )
            if rc != mqtt.MQTT_ERR_SUCCESS:
                raise RoborockException(f"Failed to connect (rc:{rc})")
        return rc == mqtt.MQTT_ERR_SUCCESS

    async def _async_response(self, request_id: int, protocol_id: int = 0) -> tuple[Any, VacuumError | None]:
        try:
            queue = RoborockQueue(protocol_id)
            self._waiting_queue[request_id] = queue
            (response, err) = await queue.async_get(QUEUE_TIMEOUT)
            return response, err
        except (TimeoutError, CancelledError):
            raise RoborockTimeout(
                f"Timeout after {QUEUE_TIMEOUT} seconds waiting for response"
            ) from None
        finally:
            del self._waiting_queue[request_id]

    async def async_disconnect(self) -> Any:
        async with self._mutex:
            disconnecting = self.sync_disconnect()
            if disconnecting:
                (response, err) = await self._async_response(1)
                if err:
                    raise RoborockException(err) from err
                return response

    async def async_connect(self) -> Any:
        async with self._mutex:
            connecting = self.sync_connect()
            if connecting:
                (response, err) = await self._async_response(0)
                if err:
                    raise RoborockException(err) from err
                return response

    async def validate_connection(self) -> None:
        await self.async_connect()

    def _decode_msg(self, msg, device: HomeDataDevice) -> dict[str, Any]:
        if msg[0:3] != "1.0".encode():
            raise RoborockException("Unknown protocol version")
        crc32 = binascii.crc32(msg[0: len(msg) - 4])
        expected_crc32 = struct.unpack_from("!I", msg, len(msg) - 4)
        if crc32 != expected_crc32[0]:
            raise RoborockException(f"Wrong CRC32 {crc32}, expected {expected_crc32}")

        [version, _seq, _random, timestamp, protocol, payload_len] = struct.unpack(
            "!3sIIIHH", msg[0:19]
        )
        [payload, expected_crc32] = struct.unpack_from(f"!{payload_len}sI", msg, 19)
        if crc32 != expected_crc32:
            raise RoborockException(f"Wrong CRC32 {crc32}, expected {expected_crc32}")

        aes_key = md5bin(encode_timestamp(timestamp) + device.local_key + self._salt)
        decipher = AES.new(aes_key, AES.MODE_ECB)
        decrypted_payload = unpad(decipher.decrypt(payload), AES.block_size)
        return {
            "version": version,
            "timestamp": timestamp,
            "protocol": protocol,
            "payload": decrypted_payload,
        }

    def _send_msg_raw(self, device_id, protocol, timestamp, payload) -> None:
        local_key = self.device_map[device_id].device.local_key
        aes_key = md5bin(encode_timestamp(timestamp) + local_key + self._salt)
        cipher = AES.new(aes_key, AES.MODE_ECB)
        encrypted = cipher.encrypt(pad(payload, AES.block_size))
        msg = struct.pack(
            "!3sIIIHH",
            "1.0".encode(),
            self._seq,
            self._random,
            timestamp,
            protocol,
            len(encrypted),
        )
        msg = msg[0:19] + encrypted
        crc32 = binascii.crc32(msg)
        msg += struct.pack("!I", crc32)
        info = self.publish(
            f"rr/m/i/{self._mqtt_user}/{self._hashed_user}/{device_id}", msg
        )
        if info.rc != mqtt.MQTT_ERR_SUCCESS:
            raise RoborockException(f"Failed to publish (rc: {info.rc})")

    async def send_command(
            self, device_id: str, method: RoborockCommand, params: list = None
    ):
        await self.validate_connection()
        timestamp = math.floor(time.time())
        request_id = self._id_counter
        self._id_counter = (self._id_counter + 1) % 32767
        inner = {
            "id": request_id,
            "method": method,
            "params": params or [],
            "security": {
                "endpoint": self._endpoint,
                "nonce": self._nonce.hex().upper(),
            },
        }
        payload = bytes(
            json.dumps(
                {
                    "t": timestamp,
                    "dps": {"101": json.dumps(inner, separators=(",", ":"))},
                },
                separators=(",", ":"),
            ).encode()
        )
        _LOGGER.debug(f"id={request_id} Requesting method {method} with {params}")
        request_protocol = 101
        response_protocol = 301 if method in COMMANDS_WITH_BINARY_RESPONSE else 102
        self._send_msg_raw(device_id, request_protocol, timestamp, payload)
        (response, err) = await self._async_response(request_id, response_protocol)
        if err:
            raise CommandVacuumError(method, err) from err
        if response_protocol == 301:
            _LOGGER.debug(
                f"id={request_id} Response from {method}: {len(response)} bytes"
            )
        else:
            _LOGGER.debug(f"id={request_id} Response from {method}: {response}")
        return response

    async def get_status(self, device_id: str) -> Status:
        status = await self.send_command(device_id, RoborockCommand.GET_STATUS)
        if isinstance(status, dict):
            return Status(status)

    async def get_dnd_timer(self, device_id: str) -> DNDTimer:
        dnd_timer = await self.send_command(device_id, RoborockCommand.GET_DND_TIMER)
        if isinstance(dnd_timer, dict):
            return DNDTimer(dnd_timer)

    async def get_clean_summary(self, device_id: str) -> CleanSummary:
        clean_summary = await self.send_command(
            device_id, RoborockCommand.GET_CLEAN_SUMMARY
        )
        if isinstance(clean_summary, dict):
            return CleanSummary(clean_summary)
        elif isinstance(clean_summary, bytes):
            return CleanSummary({"clean_time": clean_summary})

    async def get_clean_record(self, device_id: str, record_id: int) -> CleanRecord:
        clean_record = await self.send_command(
            device_id, RoborockCommand.GET_CLEAN_RECORD, [record_id]
        )
        if isinstance(clean_record, dict):
            return CleanRecord(clean_record)

    async def get_consumable(self, device_id: str) -> Consumable:
        consumable = await self.send_command(device_id, RoborockCommand.GET_CONSUMABLE)
        if isinstance(consumable, dict):
            return Consumable(consumable)

    async def get_prop(self, device_id: str) -> RoborockDeviceProp:
        [status, dnd_timer, clean_summary, consumable] = await asyncio.gather(
            *[
                self.get_status(device_id),
                self.get_dnd_timer(device_id),
                self.get_clean_summary(device_id),
                self.get_consumable(device_id),
            ]
        )
        last_clean_record = None
        if clean_summary and clean_summary.records and len(clean_summary.records) > 0:
            last_clean_record = await self.get_clean_record(
                device_id, clean_summary.records[0]
            )
        if any([status, dnd_timer, clean_summary, consumable]):
            return RoborockDeviceProp(
                status, dnd_timer, clean_summary, consumable, last_clean_record
            )

    async def get_multi_maps_list(self, device_id) -> MultiMapsList:
        multi_maps_list = await self.send_command(
            device_id, RoborockCommand.GET_MULTI_MAPS_LIST
        )
        if isinstance(multi_maps_list, dict):
            return MultiMapsList(multi_maps_list)

    async def get_map_v1(self, device_id):
        return await self.send_command(device_id, RoborockCommand.GET_MAP_V1)


class RoborockClient:
    def __init__(self, username: str, base_url=None) -> None:
        """Sample API Client."""
        self._username = username
        self._default_url = "https://euiot.roborock.com"
        self.base_url = base_url
        self._device_identifier = secrets.token_urlsafe(16)

    async def _get_base_url(self) -> str:
        if not self.base_url:
            url_request = PreparedRequest(self._default_url)
            response = await url_request.request(
                "post",
                "/api/v1/getUrlByEmail",
                params={"email": self._username, "needtwostepauth": "false"},
            )
            if response.get("code") != 200:
                raise RoborockException(response.get("error"))
            self.base_url = response.get("data").get("url")
        return self.base_url

    def _get_header_client_id(self):
        md5 = hashlib.md5()
        md5.update(self._username.encode())
        md5.update(self._device_identifier.encode())
        return base64.b64encode(md5.digest()).decode()

    async def request_code(self) -> None:
        base_url = await self._get_base_url()
        header_clientid = self._get_header_client_id()
        code_request = PreparedRequest(base_url, {"header_clientid": header_clientid})

        code_response = await code_request.request(
            "post",
            "/api/v1/sendEmailCode",
            params={
                "username": self._username,
                "type": "auth",
            },
        )

        if code_response.get("code") != 200:
            raise RoborockException(code_response.get("msg"))

    async def pass_login(self, password: str) -> UserData:
        base_url = await self._get_base_url()
        header_clientid = self._get_header_client_id()

        login_request = PreparedRequest(base_url, {"header_clientid": header_clientid})
        login_response = await login_request.request(
            "post",
            "/api/v1/login",
            params={
                "username": self._username,
                "password": password,
                "needtwostepauth": "false",
            },
        )

        if login_response.get("code") != 200:
            raise RoborockException(login_response.get("msg"))
        return UserData(login_response.get("data"))

    async def code_login(self, code) -> UserData:
        base_url = await self._get_base_url()
        header_clientid = self._get_header_client_id()

        login_request = PreparedRequest(base_url, {"header_clientid": header_clientid})
        login_response = await login_request.request(
            "post",
            "/api/v1/loginWithCode",
            params={
                "username": self._username,
                "verifycode": code,
                "verifycodetype": "AUTH_EMAIL_CODE",
            },
        )

        if login_response.get("code") != 200:
            raise RoborockException(login_response.get("msg"))
        return UserData(login_response.get("data"))

    async def get_home_data(self, user_data: UserData) -> HomeData:
        base_url = await self._get_base_url()
        header_clientid = self._get_header_client_id()
        rriot = user_data.rriot
        home_id_request = PreparedRequest(
            base_url, {"header_clientid": header_clientid}
        )
        home_id_response = await home_id_request.request(
            "get",
            "/api/v1/getHomeDetail",
            headers={"Authorization": user_data.token},
        )
        if home_id_response.get("code") != 200:
            raise RoborockException(home_id_response.get("msg"))
        home_id = home_id_response.get("data").get("rrHomeId")
        timestamp = math.floor(time.time())
        nonce = secrets.token_urlsafe(6)
        prestr = ":".join(
            [
                rriot.user,
                rriot.password,
                nonce,
                str(timestamp),
                hashlib.md5(("/user/homes/" + str(home_id)).encode()).hexdigest(),
                "",
                "",
            ]
        )
        mac = base64.b64encode(
            hmac.new(rriot.h_unknown.encode(), prestr.encode(), hashlib.sha256).digest()
        ).decode()
        home_request = PreparedRequest(
            rriot.reference.api,
            {
                "Authorization": f'Hawk id="{rriot.user}", s="{rriot.password}", ts="{timestamp}", nonce="{nonce}", '
                                 f'mac="{mac}"',
            },
        )
        home_response = await home_request.request("get", "/user/homes/" + str(home_id))
        if not home_response.get("success"):
            raise RoborockException(home_response)
        home_data = home_response.get("result")
        return HomeData(home_data)



from enum import Enum

from roborock.containers import HomeDataDevice, HomeDataProduct, Status, CleanSummary, Consumable, \
    DNDTimer, CleanRecord


class RoborockDevicePropField(str, Enum):
    STATUS = "status"
    DND_TIMER = "dnd_timer"
    CLEAN_SUMMARY = "clean_summary"
    CONSUMABLE = "consumable"
    LAST_CLEAN_RECORD = "last_clean_record"

class RoborockCommand(str, Enum):
    GET_MAP_V1 = "get_map_v1",
    GET_STATUS = "get_status"
    GET_DND_TIMER = "get_dnd_timer"
    GET_CLEAN_SUMMARY = "get_clean_summary"
    GET_CLEAN_RECORD = "get_clean_record"
    GET_CONSUMABLE = "get_consumable"
    GET_MULTI_MAPS_LIST = "get_multi_maps_list",
    APP_START = "app_start",
    APP_PAUSE = "app_pause",
    APP_STOP = "app_stop",
    APP_CHARGE = "app_charge",
    APP_SPOT = "app_spot",
    FIND_ME = "find_me",
    SET_CUSTOM_MODE = "set_custom_mode",
    SET_MOP_MODE = "set_mop_mode",
    SET_WATER_BOX_CUSTOM_MODE = "set_water_box_custom_mode",
    RESET_CONSUMABLE = "reset_consumable",
    LOAD_MULTI_MAP = "load_multi_map",
    APP_RC_START = "app_rc_start",
    APP_RC_END = "app_rc_end",
    APP_RC_MOVE = "app_rc_move",
    APP_GOTO_TARGET = "app_goto_target",
    APP_SEGMENT_CLEAN = "app_segment_clean",
    APP_ZONED_CLEAN = "app_zoned_clean",


class RoborockDeviceInfo:
    def __init__(self, device: HomeDataDevice, product: HomeDataProduct):
        self.device = device
        self.product = product


class RoborockDeviceProp:
    def __init__(self, status: Status, dnd_timer: DNDTimer, clean_summary: CleanSummary, consumable: Consumable,
                 last_clean_record: CleanRecord):
        self.status = status
        self.dnd_timer = dnd_timer
        self.clean_summary = clean_summary
        self.consumable = consumable
        self.last_clean_record = last_clean_record

    def update(self, device_prop: 'RoborockDeviceProp'):
        if device_prop.status:
            self.status = device_prop.status
        if device_prop.dnd_timer:
            self.dnd_timer = device_prop.dnd_timer
        if device_prop.clean_summary:
            self.clean_summary = device_prop.clean_summary
        if device_prop.consumable:
            self.consumable = device_prop.consumable
        if device_prop.last_clean_record:
            self.last_clean_record = device_prop.last_clean_record
